@@ -1,4 +1,5 @@
-# Copyright (C) 2002-2013, Stefan Schwarzer <sschwarzer@sschwarzer.net>
+# Copyright (C) 2002-2014, Stefan Schwarzer <sschwarzer@sschwarzer.net>
+# and ftputil contributors (see `doc/contributors.txt`)
 # See the file LICENSE for licensing terms.
 
 """
@@ -8,6 +9,7 @@ ftputil.stat - stat result, parsers, and FTP stat'ing for `ftputil`
 from __future__ import absolute_import
 from __future__ import unicode_literals
 
+import datetime
 import math
 import re
 import stat
@@ -21,6 +23,12 @@ import ftputil.stat_cache
 __all__ = ["StatResult", "Parser", "UnixParser", "MSParser"]
 
 
+# Datetime precision values in seconds.
+MINUTE_PRECISION  = 60
+DAY_PRECISION     = 24 * 60 * 60
+UNKNOWN_PRECISION = None
+
+
 class StatResult(tuple):
     """
     Support class resembling a tuple like that returned from
@@ -46,7 +54,7 @@ class StatResult(tuple):
         # These may be overwritten in a `Parser.parse_line` method.
         self._st_name = ""
         self._st_target = None
-        self._st_mtime_precision = None
+        self._st_mtime_precision = UNKNOWN_PRECISION
 
     def __getattr__(self, attr_name):
         if attr_name in self._index_mapping:
@@ -147,6 +155,55 @@ class Parser(object):
                   "unknown file type character '{0}'".format(file_type))
         return st_mode
 
+    def _as_int(self, int_string, int_description):
+        """
+        Return `int_string` converted to an integer.
+
+        If it can't be converted, raise a `ParserError`, using
+        `int_description` in the error message. For example, if the
+        integer value is a day, pass "day" for `int_description`.
+        """
+        try:
+            return int(int_string)
+        except ValueError:
+            raise ftputil.error.ParserError("non-integer {0} value {1!r}".
+                                            format(int_description,
+                                                   int_string))
+
+    def _mktime(self, mktime_tuple):
+        """
+        Return a float value like `time.mktime` does, but ...
+
+        - Raise a `ParserError` if parts of `mktime_tuple` are
+          invalid (say, a day is 32).
+
+        - If the resulting float value would be smaller than 0.0
+          (indicating a time before the "epoch") return a sentinel
+          value of 0.0. Do this also if the native `mktime`
+          implementation would raise an `OverflowError`.
+        """
+        datetime_tuple = mktime_tuple[:6]
+        try:
+            # Only for sanity checks, we're not interested in the
+            # return value.
+            datetime.datetime(*datetime_tuple)
+        # For example, day == 32. Not all implementations of `mktime`
+        # catch this kind of error.
+        except ValueError:
+            invalid_datetime = ("%04d-%02d-%02d %02d:%02d:%02d" %
+                                datetime_tuple)
+            raise ftputil.error.ParserError("invalid datetime {0!r}".
+                                            format(invalid_datetime))
+        try:
+            time_float = time.mktime(mktime_tuple)
+        except (OverflowError, ValueError):
+            # Sentinel for times before the epoch, see ticket #83.
+            time_float = 0.0
+        # Don't allow float values smaller than 0.0 because, according
+        # to https://docs.python.org/3/library/time.html#module-time ,
+        # these might be undefined for some platforms.
+        return max(0.0, time_float)
+
     def parse_unix_time(self, month_abbreviation, day, year_or_time,
                         time_shift, with_precision=False):
         """
@@ -159,7 +216,10 @@ class Parser(object):
         If `with_precision` is true (default: false), return a
         two-element tuple consisting of the floating point number as
         described in the previous paragraph and the precision of the
-        time in seconds. This takes into account that, for example, a
+        time in seconds. The default is `False` for backward
+        compatibility with custom parsers.
+
+        The precision value takes into account that, for example, a
         time string like "May 26  2005" has only a precision of one
         day. This information is important for the `upload_if_newer`
         and `download_if_newer` methods in the `FTPHost` class.
@@ -171,32 +231,31 @@ class Parser(object):
 
         - "May 26  2005" (month name, day of month, year)
 
-        If this method can not make sense of the given arguments, it
+        If this method can't make sense of the given arguments, it
         raises an `ftputil.error.ParserError`.
         """
         try:
             month = self._month_numbers[month_abbreviation.lower()]
         except KeyError:
-            raise ftputil.error.ParserError("invalid month abbreviation '{0}'".
+            raise ftputil.error.ParserError("invalid month abbreviation {0!r}".
                                             format(month_abbreviation))
-        day = int(day)
+        day = self._as_int(day, "day")
         if ":" not in year_or_time:
             # `year_or_time` is really a year.
-            year, hour, minute = int(year_or_time), 0, 0
-            st_mtime = time.mktime( (year, month, day,
-                                     hour, minute, 0, 0, 0, -1) )
-            # Precise up to a day.
-            st_mtime_precision = 24 * 60 * 60
+            year, hour, minute = self._as_int(year_or_time, "year"), 0, 0
+            st_mtime = self._mktime( (year, month, day,
+                                      hour, minute, 0, 0, 0, -1) )
+            st_mtime_precision = DAY_PRECISION
         else:
             # `year_or_time` is a time hh:mm.
             hour, minute = year_or_time.split(":")
-            year, hour, minute = None, int(hour), int(minute)
+            year, hour, minute = (
+              None, self._as_int(hour, "hour"), self._as_int(minute, "minute"))
             # Try the current year
             year = time.localtime()[0]
-            st_mtime = time.mktime( (year, month, day,
-                                     hour, minute, 0, 0, 0, -1) )
-            # Times are precise up to a minute.
-            st_mtime_precision = 60
+            st_mtime = self._mktime( (year, month, day,
+                                      hour, minute, 0, 0, 0, -1) )
+            st_mtime_precision = MINUTE_PRECISION
             # Rhs of comparison: Transform client time to server time
             # (as on the lhs), so both can be compared with respect
             # to the set time shift (see the definition of the time
@@ -204,15 +263,20 @@ class Parser(object):
             # last addend allows for small deviations between the
             # supposed (rounded) and the actual time shift.
             #
-            # #XXX The downside of this "correction" is that there is
+            # XXX The downside of this "correction" is that there is
             # a one-minute time interval exactly one year ago that
             # may cause that datetime to be recognized as the current
             # datetime, but after all the datetime from the server
             # can only be exact up to a minute.
             if st_mtime > time.time() + time_shift + st_mtime_precision:
                 # If it's in the future, use previous year.
-                st_mtime = time.mktime( (year-1, month, day,
-                                         hour, minute, 0, 0, 0, -1) )
+                st_mtime = self._mktime( (year-1, month, day,
+                                          hour, minute, 0, 0, 0, -1) )
+        # If we had a datetime before the epoch, the resulting value
+        # 0.0 doesn't tell us anything about the precision.
+        if st_mtime == 0.0:
+            st_mtime_precision = UNKNOWN_PRECISION
+        #
         if with_precision:
             return st_mtime, st_mtime_precision
         else:
@@ -233,7 +297,7 @@ class Parser(object):
         format "10-23-01 03:25PM" (month-day_of_month-two_digit_year,
         hour:minute, am/pm).
 
-        If this method can not make sense of the given arguments, it
+        If this method can't make sense of the given arguments, it
         raises an `ftputil.error.ParserError`.
         """
         # Derived classes might want to use `self`.
@@ -248,26 +312,28 @@ class Parser(object):
         # `MSParser.parse_line`. Should you find yourself needing
         # support for `with_precision` for a derived class, please
         # send a mail (see ftputil.txt/html).
+        month, day, year = [self._as_int(part, "year/month/day")
+                            for part in date.split("-")]
+        if year >= 1000:
+            # We have a four-digit year, so no need for heuristics.
+            pass
+        elif year >= 70:
+            year = 1900 + year
+        else:
+            year = 2000 + year
         try:
-            month, day, year = [int(part) for part in date.split("-")]
-            if year >= 1000:
-                # We have a four-digit year, so no need for heuristics.
-                pass
-            elif year >= 70:
-                year = 1900 + year
-            else:
-                year = 2000 + year
             hour, minute, am_pm = time_[0:2], time_[3:5], time_[5]
-            hour, minute = int(hour), int(minute)
-        except (ValueError, IndexError):
+        except IndexError:
             raise ftputil.error.ParserError("invalid time string '{0}'".
                                             format(time_))
+        hour, minute = (
+          self._as_int(hour, "hour"), self._as_int(minute, "minute"))
         if hour == 12 and am_pm == "A":
             hour = 0
         if hour != 12 and am_pm == "P":
             hour += 12
-        st_mtime = time.mktime( (year, month, day,
-                                 hour, minute, 0, 0, 0, -1) )
+        st_mtime = self._mktime( (year, month, day,
+                                  hour, minute, 0, 0, 0, -1) )
         return st_mtime
 
 
@@ -374,7 +440,7 @@ class MSParser(Parser):
         If the line can't be parsed, raise a `ParserError`.
 
         The parameter `time_shift` isn't used in this method but is
-        listed for compatibilty with the base class.
+        listed for compatibility with the base class.
         """
         # The local variables are rather simple.
         # pylint: disable=too-many-locals
@@ -422,7 +488,12 @@ class MSParser(Parser):
         stat_result._st_name = name
         stat_result._st_target = None
         # mtime precision in seconds
-        stat_result._st_mtime_precision = 60
+        #  If we had a datetime before the epoch, the resulting value
+        #  0.0 doesn't tell us anything about the precision.
+        if st_mtime == 0.0:
+            stat_result._st_mtime_precision = UNKNOWN_PRECISION
+        else:
+            stat_result._st_mtime_precision = MINUTE_PRECISION
         return stat_result
 
 #
