@@ -1,4 +1,5 @@
-# Copyright (C) 2003-2013, Stefan Schwarzer <sschwarzer@sschwarzer.net>
+# Copyright (C) 2003-2014, Stefan Schwarzer <sschwarzer@sschwarzer.net>
+# and ftputil contributors (see `doc/contributors.txt`)
 # See the file LICENSE for licensing terms.
 
 from __future__ import absolute_import
@@ -12,6 +13,7 @@ import unittest
 import ftputil
 import ftputil.error
 import ftputil.stat
+from ftputil.stat import MINUTE_PRECISION, DAY_PRECISION, UNKNOWN_PRECISION
 
 from test import test_base
 from test import mock_ftplib
@@ -26,14 +28,22 @@ def _test_stat(session_factory):
     return stat
 
 
+# Special value to handle special case of datetimes before the epoch.
+EPOCH = time.gmtime(0)[:6]
+
 def stat_tuple_to_seconds(t):
     """
     Return a float number representing the local time associated with
     the six-element tuple `t`.
     """
     assert len(t) == 6, \
-           "need a six-element tuple (year, month, day, hour, min, sec)"
-    return time.mktime(t + (0, 0, -1))
+             "need a six-element tuple (year, month, day, hour, min, sec)"
+    # Do _not_ apply `time.mktime` to the `EPOCH` value below. On some
+    # platforms (e. g. Windows) this might cause an `OverflowError`.
+    if t == EPOCH:
+        return 0.0
+    else:
+        return time.mktime(t + (0, 0, -1))
 
 
 class TestParsers(unittest.TestCase):
@@ -46,8 +56,10 @@ class TestParsers(unittest.TestCase):
         for line, expected_stat_result in zip(lines, expected_stat_results):
             # Convert to list to compare with the list `expected_stat_results`.
             parse_result = parser.parse_line(line)
-            stat_result = list(parse_result) + [parse_result._st_name,
-                                                parse_result._st_target]
+            stat_result = list(parse_result) + \
+                          [parse_result._st_mtime_precision,
+                           parse_result._st_name,
+                           parse_result._st_target]
             # Convert time tuple to seconds.
             expected_stat_result[8] = \
               stat_tuple_to_seconds(expected_stat_result[8])
@@ -88,20 +100,76 @@ class TestParsers(unittest.TestCase):
           # The year value for this line will change with the actual time.
           "-rw-r--r--   1 45854    200          4604 Dec 19 23:11 index.html",
           "drwxr-sr-x   2 45854    200           512 May 29  2000 os2",
+          "----------   2 45854    200           512 May 29  2000 some_file",
           "lrwxrwxrwx   2 45854    200           512 May 29  2000 osup -> "
                                                                   "../os2"
-          ]
+        ]
         expected_stat_results = [
           [17901, None, None, 2, "45854", "200", 512, None,
-           (2000, 5, 4, 0, 0, 0), None, "chemeng link", "chemeng target"],
+           (2000, 5, 4, 0, 0, 0), None, DAY_PRECISION,
+           "chemeng link", "chemeng target"],
           [33188, None, None, 1, "45854", "200", 4604, None,
-           (self._expected_year(), 12, 19, 23, 11, 0), None,
+           (self._expected_year(), 12, 19, 23, 11, 0), None, MINUTE_PRECISION,
            "index.html", None],
           [17901, None, None, 2, "45854", "200", 512, None,
-           (2000, 5, 29, 0, 0, 0), None, "os2", None],
+           (2000, 5, 29, 0, 0, 0), None, DAY_PRECISION,
+           "os2", None],
+          [32768, None, None, 2, "45854", "200", 512, None,
+           (2000, 5, 29, 0, 0, 0), None, DAY_PRECISION,
+           "some_file", None],
           [41471, None, None, 2, "45854", "200", 512, None,
-           (2000, 5, 29, 0, 0, 0), None, "osup", "../os2"]
-          ]
+           (2000, 5, 29, 0, 0, 0), None, DAY_PRECISION,
+           "osup", "../os2"]
+        ]
+        self._test_valid_lines(ftputil.stat.UnixParser, lines,
+                               expected_stat_results)
+
+    def test_alternative_unix_format(self):
+        # See http://ftputil.sschwarzer.net/trac/ticket/12 for a
+        # description for the need for an alternative format.
+        lines = [
+          "drwxr-sr-x   2   200           512 May  4  2000 "
+            "chemeng link -> chemeng target",
+          # The year value for this line will change with the actual time.
+          "-rw-r--r--   1   200          4604 Dec 19 23:11 index.html",
+          "drwxr-sr-x   2   200           512 May 29  2000 os2",
+          "lrwxrwxrwx   2   200           512 May 29  2000 osup -> ../os2"
+        ]
+        expected_stat_results = [
+          [17901, None, None, 2, None, "200", 512, None,
+           (2000, 5, 4, 0, 0, 0), None, DAY_PRECISION,
+           "chemeng link", "chemeng target"],
+          [33188, None, None, 1, None, "200", 4604, None,
+           (self._expected_year(), 12, 19, 23, 11, 0), None, MINUTE_PRECISION,
+           "index.html", None],
+          [17901, None, None, 2, None, "200", 512, None,
+           (2000, 5, 29, 0, 0, 0), None, DAY_PRECISION,
+           "os2", None],
+          [41471, None, None, 2, None, "200", 512, None,
+           (2000, 5, 29, 0, 0, 0), None, DAY_PRECISION,
+           "osup", "../os2"]
+        ]
+        self._test_valid_lines(ftputil.stat.UnixParser, lines,
+                               expected_stat_results)
+
+    def test_pre_epoch_times_for_unix(self):
+        # See http://ftputil.sschwarzer.net/trac/ticket/83 .
+        # `mirrors.ibiblio.org` returns dates before the "epoch" that
+        # cause an `OverflowError` in `mktime` on some platforms,
+        # e. g. Windows.
+        lines = [
+          "-rw-r--r--   1 45854    200          4604 May  4  1968 index.html",
+          "-rw-r--r--   1 45854    200          4604 Dec 31  1969 index.html",
+          "-rw-r--r--   1 45854    200          4604 May  4  1800 index.html",
+        ]
+        expected_stat_result = \
+          [33188, None, None, 1, "45854", "200", 4604, None,
+           EPOCH, None, UNKNOWN_PRECISION, "index.html", None]
+        # Make shallow copies to avoid converting the time tuple more
+        # than once in _test_valid_lines`.
+        expected_stat_results = [expected_stat_result[:],
+                                 expected_stat_result[:],
+                                 expected_stat_result[:]]
         self._test_valid_lines(ftputil.stat.UnixParser, lines,
                                expected_stat_results)
 
@@ -112,6 +180,16 @@ class TestParsers(unittest.TestCase):
           "total 14",
           # Invalid month abbreviation
           "drwxr-sr-x   2 45854    200           512 Max  4  2000 chemeng",
+          # Year value isn't an integer
+          "drwxr-sr-x   2 45854    200           512 May  4  abcd chemeng",
+          # Day value isn't an integer
+          "drwxr-sr-x   2 45854    200           512 May ab  2000 chemeng",
+          # Hour value isn't an integer
+          "-rw-r--r--   1 45854    200          4604 Dec 19 ab:11 index.html",
+          # Minute value isn't an integer
+          "-rw-r--r--   1 45854    200          4604 Dec 19 23:ab index.html",
+          # Day value too large
+          "drwxr-sr-x   2 45854    200           512 May 32  2000 chemeng",
           # Incomplete mode
           "drwxr-sr-    2 45854    200           512 May  4  2000 chemeng",
           # Invalid first letter in mode
@@ -125,34 +203,9 @@ class TestParsers(unittest.TestCase):
             "os1 -> os2 -> os3",
           # Missing name
           "-rwxr-sr-x   2 45854    200           51x May  4  2000 ",
-          ]
+        ]
         self._test_invalid_lines(ftputil.stat.UnixParser, lines)
 
-    def test_alternative_unix_format(self):
-        # See http://ftputil.sschwarzer.net/trac/ticket/12 for a
-        # description for the need for an alternative format.
-        lines = [
-          "drwxr-sr-x   2   200           512 May  4  2000 "
-            "chemeng link -> chemeng target",
-          # The year value for this line will change with the actual time.
-          "-rw-r--r--   1   200          4604 Dec 19 23:11 index.html",
-          "drwxr-sr-x   2   200           512 May 29  2000 os2",
-          "lrwxrwxrwx   2   200           512 May 29  2000 osup -> ../os2"
-          ]
-        expected_stat_results = [
-          [17901, None, None, 2, None, "200", 512, None,
-           (2000, 5, 4, 0, 0, 0), None, "chemeng link", "chemeng target"],
-          [33188, None, None, 1, None, "200", 4604, None,
-           (self._expected_year(), 12, 19, 23, 11, 0), None,
-           "index.html", None],
-          [17901, None, None, 2, None, "200", 512, None,
-           (2000, 5, 29, 0, 0, 0), None, "os2", None],
-          [41471, None, None, 2, None, "200", 512, None,
-           (2000, 5, 29, 0, 0, 0), None, "osup", "../os2"]
-          ]
-        self._test_valid_lines(ftputil.stat.UnixParser, lines,
-                               expected_stat_results)
-
     #
     # Microsoft parser
     #
@@ -163,19 +216,24 @@ class TestParsers(unittest.TestCase):
           "07-17-00  02:08PM             12266720 test.exe",
           "07-17-09  12:08AM             12266720 test.exe",
           "07-17-09  12:08PM             12266720 test.exe"
-          ]
+        ]
         expected_stat_results = [
           [16640, None, None, None, None, None, None, None,
-           (2001, 7, 27, 11, 16, 0), None, "Test", None],
+           (2001, 7, 27, 11, 16, 0), None, MINUTE_PRECISION,
+           "Test", None],
           [16640, None, None, None, None, None, None, None,
-           (1995, 10, 23, 15, 25, 0), None, "WindowsXP", None],
+           (1995, 10, 23, 15, 25, 0), None, MINUTE_PRECISION,
+           "WindowsXP", None],
           [33024, None, None, None, None, None, 12266720, None,
-           (2000, 7, 17, 14, 8, 0), None, "test.exe", None],
+           (2000, 7, 17, 14, 8, 0), None, MINUTE_PRECISION,
+           "test.exe", None],
           [33024, None, None, None, None, None, 12266720, None,
-           (2009, 7, 17, 0, 8, 0), None, "test.exe", None],
+           (2009, 7, 17, 0, 8, 0), None, MINUTE_PRECISION,
+           "test.exe", None],
           [33024, None, None, None, None, None, 12266720, None,
-           (2009, 7, 17, 12, 8, 0), None, "test.exe", None]
-          ]
+           (2009, 7, 17, 12, 8, 0), None, MINUTE_PRECISION,
+           "test.exe", None]
+        ]
         self._test_valid_lines(ftputil.stat.MSParser, lines,
                                expected_stat_results)
 
@@ -183,14 +241,20 @@ class TestParsers(unittest.TestCase):
         # See http://ftputil.sschwarzer.net/trac/ticket/67
         lines = [
           "10-19-2012  03:13PM       <DIR>          SYNCDEST",
-          "10-19-2012  03:13PM       <DIR>          SYNCSOURCE"
-          ]
+          "10-19-2012  03:13PM       <DIR>          SYNCSOURCE",
+          "10-19-1968  03:13PM       <DIR>          SYNC"
+        ]
         expected_stat_results = [
           [16640, None, None, None, None, None, None, None,
-           (2012, 10, 19, 15, 13, 0), None, "SYNCDEST", None],
+           (2012, 10, 19, 15, 13, 0), None, MINUTE_PRECISION,
+           "SYNCDEST", None],
           [16640, None, None, None, None, None, None, None,
-           (2012, 10, 19, 15, 13, 0), None, "SYNCSOURCE", None],
-          ]
+           (2012, 10, 19, 15, 13, 0), None, MINUTE_PRECISION,
+           "SYNCSOURCE", None],
+          [16640, None, None, None, None, None, None, None,
+           EPOCH, None, UNKNOWN_PRECISION,
+           "SYNC", None],
+        ]
         self._test_valid_lines(ftputil.stat.MSParser, lines,
                                expected_stat_results)
 
@@ -200,9 +264,17 @@ class TestParsers(unittest.TestCase):
           "07-27-01  11:16AM                      Test",
           # "AM"/"PM" missing
           "07-17-00  02:08             12266720 test.exe",
+          # Year not an int
+          "07-17-ab  02:08AM           12266720 test.exe",
+          # Month not an int
+          "ab-17-00  02:08AM           12266720 test.exe",
+          # Day not an int
+          "07-ab-00  02:08AM           12266720 test.exe",
+          # Hour not an int
+          "07-17-00  ab:08AM           12266720 test.exe",
           # Invalid size value
           "07-17-00  02:08AM           1226672x test.exe"
-          ]
+        ]
         self._test_invalid_lines(ftputil.stat.MSParser, lines)
 
     #
